using System;
using System.Collections.Generic;
using System.Text;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.Data.OleDb;
using System.Threading;
using System.Globalization;

namespace SqlBi.Tools.SqlBulkTool {

    /// <summary>
    /// 
    /// Database copy mantains connection information about source and destination database and 
    /// delivers SqlConnections to its tables.
    /// 
    /// Upon construction and properly configuration it is only necessary to launch LoadAllTables
    /// to have all the configured tables copied from source to destination
    /// 
    /// Each partition in each table is loaded in a separate thread. 
    /// 
    /// This class has two abstract method: 
    ///     - GetSourceConn
    ///     - GetDbCommand
    /// any implementation of DataBaseCopy must provide these two methods in order to let the system
    /// work with any implementation of DbCommand and DbConnection.
    /// 
    /// See SqlDataBaseCopy and OleDbDataBaseCopy to gather more information about real implementation
    /// of DataBaseCopy.
    /// 
    /// </summary>
    public abstract class DatabaseCopy : MarshalByRefObject {
        private string _SourceConnectionString;         // The source connection string
        private string _DestConnectionString;           // The destination connection string
        private List<TableCopy> _Tables;                // The list of tables we need to load
        private string _Name;                           // The name of the database

        public string Name {
            get { return _Name; }
            set { _Name = value; }
        }
        
        public string SourceConnectionString {
            get { return _SourceConnectionString; }
            set { _SourceConnectionString = value; }
        }

        public string DestConnectionString {
            get { return _DestConnectionString; }
            set { _DestConnectionString = value; }
        }

        public List<TableCopy> Tables {
            get { return _Tables; }
        }

        /// <summary>
        /// 
        /// Constructor for DataBaseCopy
        /// 
        /// </summary>
        public DatabaseCopy () {
            _Tables = new List<TableCopy> ();
        }

        /// <summary>
        /// 
        /// Returns a new DbCommand that can be used on the specified Connection.
        /// 
        /// </summary>
        /// <param name="conn">The connection to use</param>
        /// <param name="commandString">The commandString that represent the query to run</param>
        /// <returns></returns>
        public abstract IDbCommand GetDbCommand (string commandString);

        /// <summary>
        /// 
        /// Returns a SqlConnection valid for the destination database
        /// 
        /// </summary>
        /// <returns></returns>
        internal SqlConnection GetDestConn () {
            return new SqlConnection (DestConnectionString);
        }

        /// <summary>
        /// 
        /// Load all tables in the database, it starts loading ALL the partitions in ALL the tables, the threads
        /// will limit themself using LimitNumberOfThreads semaphore.
        /// 
        /// </summary>
        public void PrepareLoadingThreads () {
            foreach (TableCopy T in Tables) {
                Logger.LogInfo (String.Format (CultureInfo.InvariantCulture, "Preparing process of table {0}.{1} from {2}", Name, T.DestTableName, T.SourceTableName));
                T.ReadTableInformation ();
                T.PrepareLoadingThreads ();
            }
        }

        /// <summary>
        /// 
        /// Runs each command passed in the argument list against the destination database, the entire list is
        /// executed in a critical section for the table so that metadata is guaranteed to be consistent, beware
        /// that this methos is called by the Partition.Process method that is executed in a multithreaded environment.
        /// 
        /// </summary>
        /// <param name="Command"></param>
        internal void RunCriticalCommands (TableCopy Table, params string[] Command) {
            using (SqlConnection destConn = GetDestConn ()) {
                destConn.Open ();
                lock (Table) {
                    foreach (string s in Command) {
                        SqlCommand cmd = new SqlCommand (s, destConn);
                        cmd.ExecuteNonQuery ();
                    }
                }
            }
        }

        /// <summary>
        /// 
        /// Builds a SELECT statement that can be fired against source database to retrieve all the columns that
        /// are present in schema.
        /// 
        /// In the 
        /// 
        /// </summary>
        /// <param name="T">The destination table that will receive data</param>
        /// <param name="schema">The schema of columns that need to be retrieved</param>
        /// <param name="condition">Optional: a condition for the rows</param>
        /// <returns></returns>
        public virtual IDbCommand GetSourceCommand (TableCopy table, string condition) {
            StringBuilder SourceQuery = new StringBuilder ();
            DataTable Schema = table.GetSchema;
            SourceQuery.Append ("Select ");
            for (int i = 0; i < Schema.Rows.Count; i++) {
                SourceQuery.Append( EncapsulateSourceIdentifier( (string) Schema.Rows[i]["ColumnName"] ) );
                if (i < Schema.Rows.Count - 1) {
                    SourceQuery.Append (", ");
                }
            }
            SourceQuery.AppendFormat( " From {0} ", table.SourceTableName);

            if (!String.IsNullOrEmpty (condition)) {
                SourceQuery.AppendFormat ("Where {0} ", condition);
            }

            return GetDbCommand (SourceQuery.ToString ());
        }

 
        /// <summary>
        /// 
        /// Returns a valid identifier for the SQL dialect implemented. 
        /// The identifier is encapsulated between delimiters and 
        /// internal characters are substituted where necessary
        /// 
        /// </summary>
        /// <param name="identifier">Identifier to normalize</param>
        /// <returns></returns>
        public virtual string EncapsulateSourceIdentifier (string identifier) {
            return identifier;
        }

        /// <summary>
        /// 
        /// Returns a valid identifier for SQL Server
        /// The identifier is encapsulated between delimiters and 
        /// internal characters are substituted where necessary
        /// 
        /// </summary>
        /// <param name="identifier">Identifier to normalize</param>
        /// <returns></returns>
        public string EncapsulateDestinationIdentifier (string identifier) {
            return String.Format ("[{0}]", identifier.Replace ("]", "]]"));
        }
    }
}
