using System;
using System.Collections.Generic;
using System.Text;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using System.Data.OleDb;
using System.Threading;
using System.Globalization;

namespace SqlBi.Tools.SqlBulkTool {
    /// <summary>
    /// 
    /// This class is the actual implementation of a TableCopy operation
    /// 
    /// TableCopy retrieves Partition information from the system data tables of SqlServer2005 and uses
    /// them to separate the load process in partitions so that each single partition can be carried on
    /// by a separate process.
    /// 
    /// All the code here is very SQL SERVER 2005 dependant as we need to have a nice tour inside its 
    /// system tables to gather information about the partitioning.
    /// 
    /// </summary>
    public class TableCopy {
        private string _SourceTableName = "";                   // The name of the source table
        private string _DestTableName = "";                     // The name of the destination table
        private string _DestTableSchema = "";                   // The schema of the destination table
        private string _UpdateMinValue = "";                    // The minimum value to update
        private string _UpdateMaxValue = "";                    // The maximum value to update
        private string _PartitionFunctionName = "";             // The name of the partition function
        private string _PartitionColumnName = "";               // The name of the partition column
        private SqlDbType _PartitionColumnType;                 // The type of the partition column
        private bool _truncateBeforeLoad = true;                // Do we need to truncate the table before loading data?

        private bool IsPartitioned;                             // Is the table partitioned?
        private List<PartitionCopy> Partitions;                 // The list of the partitions contained in the table
        internal DatabaseCopy Database;                         // The database that owns us

        private static readonly ApplicationException BadParameterType = new ApplicationException ("Unsupported data type for column partition.");
        private static readonly ApplicationException UnknownParameterType = new ApplicationException ("Unable to determine data type for column partition.");

        public string SourceTableName {
            get { return _SourceTableName; }
            set { _SourceTableName = value; }
        }

        public string DestTableName {
            get { return _DestTableName; }
            set { _DestTableName = value; }
        }

        public string UpdateMinValue {
            get { return _UpdateMinValue; }
            set { _UpdateMinValue = value; }
        }

        public string UpdateMaxValue {
            get { return _UpdateMaxValue; }
            set { _UpdateMaxValue = value; }
        }

        public string DestTableSchema {
            get { return _DestTableSchema; }
            set { _DestTableSchema = value; }
        }
        public string PartitionFunctionName {
            get { return _PartitionFunctionName; }
            set { _PartitionFunctionName = value; }
        }
        public string PartitionColumnName {
            get { return _PartitionColumnName; }
            set { _PartitionColumnName = value; }
        }
        public SqlDbType PartitionColumnType {
            get { return _PartitionColumnType; }
            set { _PartitionColumnType = value; }
        }

        public bool TruncateBeforeLoad {
            get { return _truncateBeforeLoad; }
            set { _truncateBeforeLoad = value; }
        }
        
        /// <summary>
        /// 
        /// Standard constructors for TableCopy
        /// 
        /// </summary>
        private TableCopy () {
            Partitions = new List<PartitionCopy> ();
        }

        public TableCopy (DatabaseCopy database)
            : this () {
            Database = database;
        }

        /// <summary>
        /// 
        /// Gathers information about the source and destination table before performing the bulk copy.
        /// 
        /// The order under which these methods are called is very important, it is necessary to have
        /// Max and Min values BEFORE gathering partition information as max and min are used to determine
        /// if a partition needs rebuilding or not.
        /// 
        /// </summary>
        public void ReadTableInformation () {
            GetPartitionFunctionName ();
            if (IsPartitioned) {
                GetPartitionColumnName ();
                GetMaxAndMinValues ();
                GetPartitionInformation ();
            }
        }

        /// <summary>
        /// 
        /// Loads the source table information into the destination table, looping on each single partition and
        /// preparing an operation for each partition or for the full table, depending on the table structure
        /// 
        /// Each Operation is added to the threadManager so it will start them when necessary
        /// 
        /// </summary>
        /// <returns></returns>
        public void PrepareLoadingThreads () {
            if (IsPartitioned) {
                foreach (PartitionCopy P in Partitions) {
                    if (P.NeedsRebuild) {
                        ThreadManager.AddOperation (new PartitionOperation (P));
                    }
                }
            } else {
                ThreadManager.AddOperation (new TableOperation (this));
            }
        }

        /// <summary>
        /// 
        /// Computes max and min values of the partition column from the source table in order to decide which partitions 
        /// needs rebuilding.
        /// 
        /// The query is sent to the database ONLY if one of UpdateMinValue or UpdateMaxValue is null, it is up to the caller
        /// to decide if he/she wants to reduce loading by settinig Min/Max values to update.
        /// 
        /// </summary>
        private void GetMaxAndMinValues () {
            if ((UpdateMinValue == null) || (UpdateMaxValue == null)) {
                IDbCommand command = Database.GetDbCommand (String.Format (CultureInfo.InvariantCulture, "select Min ({0}) As First, Max ({0}) As Last from {1}",
                    Database.EncapsulateSourceIdentifier (PartitionColumnName),
                    Database.EncapsulateSourceIdentifier (SourceTableName)));

                command.Connection.Open ();
                try {
                    IDataReader reader = command.ExecuteReader ();
                    if (reader.Read ()) {
                        if (UpdateMinValue == null) {
                            UpdateMinValue = reader["First"].ToString ();
                        }
                        if (UpdateMaxValue == null) {
                            UpdateMaxValue = reader["Last"].ToString ();
                        }
                    }
                    reader.Close ();
                } finally {
                    command.Connection.Close ();
                }
            }
        }

        /// <summary>
        /// 
        /// Retrieves the partition column name from the system tables for the specified table name and schema
        /// 
        /// </summary>
        private void GetPartitionColumnName () {
            String commandString = String.Format (CultureInfo.InvariantCulture,
                @"SELECT
                    CAST(ic.partition_ordinal AS int) AS [ID],
                    c.name      AS [ParamName],
                    types.name  As [TypeName]
                FROM
                    sys.tables AS tbl
                    INNER JOIN sys.indexes AS idx 
                        ON idx.object_id = tbl.object_id and idx.index_id < 2
                    INNER JOIN sys.index_columns ic 
                        ON (ic.partition_ordinal > 0) AND (ic.index_id=idx.index_id AND ic.object_id=CAST(tbl.object_id AS int))
                    INNER JOIN sys.columns c 
                        ON c.object_id = ic.object_id and c.column_id = ic.column_id
                    Inner Join sys.types types
                        ON c.system_type_id = types.system_type_id
                WHERE
                    (tbl.name='{0}') and SCHEMA_NAME(tbl.schema_id)='{1}'
                ORDER BY
                    [ID] ASC",
                DestTableName, DestTableSchema);

            using (SqlConnection conn = Database.GetDestConn ()) {
                SqlCommand command = new SqlCommand (commandString, conn);
               
                conn.Open ();
                SqlDataReader reader = command.ExecuteReader ();
                while (reader.Read ()) {
                    string colName = reader["ParamName"].ToString ();
                    string typeName = reader["TypeName"].ToString ();

                    PartitionColumnName = colName;
                    switch (typeName.ToLower ()) {
                        case "bigint":
                            PartitionColumnType =  SqlDbType.BigInt;
                            break;
                        case "binary":
                            PartitionColumnType = SqlDbType.Binary;
                            throw BadParameterType;
                        case "bit":
                            PartitionColumnType = SqlDbType.Bit;
                            break;
                        case "char":
                            PartitionColumnType = SqlDbType.Char;
                            break;
                        case "datetime":
                            PartitionColumnType = SqlDbType.DateTime;
                            break;
                        case "decimal":
                            PartitionColumnType = SqlDbType.Decimal;
                            break;
                        case "float":
                            PartitionColumnType = SqlDbType.Float;
                            break;
                        case "image":
                            PartitionColumnType = SqlDbType.Image;
                            throw BadParameterType;
                        case "int":
                            PartitionColumnType = SqlDbType.Int;
                            break;
                        case "money":
                            PartitionColumnType = SqlDbType.Money;
                            break;
                        case "nchar":
                            PartitionColumnType = SqlDbType.NChar;
                            break;
                        case "ntext":
                            PartitionColumnType = SqlDbType.NText;
                            throw BadParameterType;
                        case "nvarchar":
                            PartitionColumnType = SqlDbType.NVarChar;
                            break;
                        case "real":
                            PartitionColumnType = SqlDbType.Real;
                            break;
                        case "uniqueidentifier":
                            PartitionColumnType = SqlDbType.UniqueIdentifier;
                            break;
                        case "smalldatetime":
                            PartitionColumnType = SqlDbType.SmallDateTime;
                            break;
                        case "smallint":
                            PartitionColumnType = SqlDbType.SmallInt;
                            break;
                        case "smallmoney":
                            PartitionColumnType = SqlDbType.SmallMoney;
                            break;
                        case "text":
                            PartitionColumnType = SqlDbType.Text;
                            throw BadParameterType;
                        case "timestamp":
                            PartitionColumnType = SqlDbType.Timestamp;
                            break;
                        case "tinyint":
                            PartitionColumnType = SqlDbType.TinyInt;
                            break;
                        case "varbinary":
                            PartitionColumnType = SqlDbType.VarBinary;
                            throw BadParameterType;
                        case "varchar":
                            PartitionColumnType = SqlDbType.VarChar;
                            break;
                        case "variant":
                            PartitionColumnType = SqlDbType.Variant;
                            throw BadParameterType;
                        case "xml":
                            PartitionColumnType = SqlDbType.Xml;
                            throw BadParameterType;
                        case "udt":
                            PartitionColumnType = SqlDbType.Udt;
                            throw BadParameterType;
                        default:
                            throw UnknownParameterType;
                    }
                }
            }
        }

        /// <summary>
        /// 
        /// Retrieves the partition function name from the system tables for the specified table name and schema
        /// 
        /// </summary>
        private void GetPartitionFunctionName () {
            String commandString = String.Format (CultureInfo.InvariantCulture,
                @"SELECT
                    CASE 
                        WHEN 'PS'=dsidx.type THEN dsidx.name 
                        ELSE N'' 
                    END AS [PartitionScheme],
                    fun.name As [PartitionFunctionName],
                    CAST(
                        CASE 
                            WHEN 'PS'=dsidx.type THEN 1 
                            ELSE 0 
                        END 
                        AS bit) 
                    AS [IsPartitioned]
                FROM
                    sys.tables AS tbl
                    INNER JOIN sys.indexes AS idx 
                        ON idx.object_id = tbl.object_id and idx.index_id < 2
                    Left Outer JOIN sys.data_spaces AS dsidx 
                        ON dsidx.data_space_id = idx.data_space_id
                    Left Outer Join sys.partition_schemes sch
                        On dsidx.data_space_id = sch.data_space_id
                    Left Outer Join sys.partition_functions fun
                        On fun.function_id = sch.function_id
                WHERE
                    (tbl.name=N'{0}' and SCHEMA_NAME(tbl.schema_id)=N'{1}')",
                DestTableName, DestTableSchema);

            using (SqlConnection conn = Database.GetDestConn ()) {
                SqlCommand command = new SqlCommand (commandString, conn);

                conn.Open ();
                SqlDataReader reader = command.ExecuteReader ();
                if (reader.Read ()) {
                    PartitionFunctionName = reader["PartitionFunctionName"].ToString ();
                    IsPartitioned = (bool) reader["IsPartitioned"];
                }
            }
        }

        /// <summary>
        /// 
        /// Gathers partition information from destination Server in order to determine the partition structure
        /// of the destination table.
        /// 
        /// As the partition information contain only the boundary, we loop on all the partitions ordered by
        /// value and store both max and min value into our Partitions, this makes all the partition processing
        /// much easier.
        /// 
        /// </summary>
        private void GetPartitionInformation () {
            SqlConnection conn = Database.GetDestConn ();
            String commandString = String.Format (CultureInfo.InvariantCulture,
                @"Select
                    boundary_id,
                    boundary_value_on_right, 
                    value
		        From sys.partition_functions fun
		            Inner Join sys.partition_range_values val
		            On val.function_id = fun.function_id
		        Where name = '{0}' 
                Order By value", PartitionFunctionName);

            SqlCommand command = new SqlCommand (commandString, conn);
            IComparable OldValue = null;

            conn.Open ();
            try {
                SqlDataReader reader = command.ExecuteReader ();
                while (reader.Read ()) {
                    PartitionCopy newPartition = new PartitionCopy (this);
                    newPartition.MinValue = OldValue;
                    newPartition.MaxValue = reader["Value"].ToString ();
                    newPartition.PartitionId = reader["boundary_id"].ToString ();
                    newPartition.BoundaryValueOnRight = (bool) reader["boundary_value_on_right"];

                    if (newPartition.ContainsDataBetween (UpdateMinValue, UpdateMaxValue)) {
                        newPartition.NeedsRebuild = true;
                    } else {
                        newPartition.NeedsRebuild = false;
                    }
                    Partitions.Add (newPartition);
                    OldValue = newPartition.MaxValue;
                }
            } finally {
                conn.Close ();
            }
        }

        /// <summary>
        /// 
        /// Returns a DataTable that repreent the schema of the destination table.
        /// 
        /// </summary>
        /// <returns></returns>
        public DataTable GetSchema {
            get {
                DataTable Schema;
                using (SqlConnection DestConn = Database.GetDestConn ()) {
                    DestConn.Open ();
                    using (SqlCommand SchemaCommand = new SqlCommand (
                        String.Format (CultureInfo.InvariantCulture, "SELECT * FROM [{0}].[{1}]", DestTableSchema, DestTableName), DestConn)) {
                        using (SqlDataReader reader = SchemaCommand.ExecuteReader (CommandBehavior.SchemaOnly)) {
                            Schema = reader.GetSchemaTable ();
                            reader.Close ();
                        }
                    }
                    DestConn.Close ();
                }
                return Schema;
            }
        }

        /// <summary>
        /// 
        /// Loads the entire table from the source to the destination using BulkCopy.
        /// This method is an alternative to the Partition.Process method for tables that are not partitioned.
        /// 
        /// </summary>
        public void ProcessNonPartitionedTable () {
            using (SqlConnection DestConn = Database.GetDestConn ()) {
                IDbCommand SourceCommand = Database.GetSourceCommand (this, "");
                SqlCommand truncate = new SqlCommand (String.Format (CultureInfo.InvariantCulture, "Truncate Table [{0}].[{1}]", DestTableSchema, DestTableName), DestConn);

                DestConn.Open ();
                if (TruncateBeforeLoad) {
                    truncate.ExecuteNonQuery ();
                }

                SourceCommand.Connection.Open ();
                try {
                    using (IDataReader SourceReader = SourceCommand.ExecuteReader ()) {
                        SqlBulkCopy bulkCopy = new SqlBulkCopy( DestConn, SqlBulkCopyOptions.TableLock, null );
                        bulkCopy.DestinationTableName = Database.EncapsulateDestinationIdentifier (DestTableSchema) + "." + Database.EncapsulateDestinationIdentifier (DestTableName);
                        bulkCopy.BatchSize = 1000;

                        bulkCopy.WriteToServer (SourceReader);
                        bulkCopy.Close ();
                        SourceReader.Close ();
                    }
                } finally {
                    SourceCommand.Connection.Close ();
                }
            }
        }
    }
}
