using System;
using System.Collections.Generic;
using System.Text;
using System.Data;
using System.Data.Common;
using System.Data.SqlClient;
using System.Data.OleDb;
using System.Threading;
using System.Globalization;

namespace SqlBi.Tools.SqlBulkTool {
    /// <summary>
    /// 
    /// Thread Manager handles all the threads that are prepared for copying the tables and partition,
    /// it mantains a list of all the thread that should run and starts them accordingly to MaxNumber.
    /// 
    /// </summary>
    public static class ThreadManager {
        /// <summary>
        /// 
        /// LockRecord is used to guarantee that a particular partition is loaded only by ONE process at a time.
        /// 
        /// If two DataBaseOperation need to load the data into the same partition OR the same table then
        /// ThreadManager will make the second one wait until the firse one finishes its work, in this way
        /// we are always sure that a specific table/partition is handled by only ONE DataBaseOperation.
        /// 
        /// </summary>
        internal class LockRecord {
            internal DatabaseOperation Operation;
            internal string DataBaseName;
            internal string Schema;
            internal string TableName;
            internal string PartitionNumber;

            public LockRecord (DatabaseOperation operation, string dataBaseName, string schema, string tableName) {
                Operation = operation;
                this.DataBaseName = dataBaseName;
                this.Schema = schema;
                this.TableName = tableName;
                this.PartitionNumber = "";
            }

            public LockRecord (DatabaseOperation operation, string dataBaseName, string schema, string tableName, string partitionNumber) {
                Operation = operation;
                this.DataBaseName = dataBaseName;
                this.Schema = schema;
                this.TableName = tableName;
                this.PartitionNumber = partitionNumber;
            }
        }

        //
        //  The list of LockRecords that will help synchronize partition loading
        //
        private static List<LockRecord> Locks = new List<LockRecord> ();

        //
        //  The list of all the threads that should run
        //
        private static List<DatabaseOperation> WaitingThreadList = new List<DatabaseOperation> ();
        //
        //  This event is Set when all the threads have finished their work
        //
        private static ManualResetEvent WorkFinished = new ManualResetEvent (false);
        //
        //  A semaphore to limit the number of concurrently running threads
        //
        private static Semaphore Sem;
        //
        //  The number of thread still running
        //
        private static int RunningThreadCount;

        //
        // The number of threads that that raised an exception
        //
        private static int _exceptionRaisedCount = 0;

        //
        // The number of threads that raised an exception
        //
        public static int ExceptionRaisedCount
        {
            get { return _exceptionRaisedCount; }
        }

        /// <summary>
        /// 
        /// This function checks if Locks contains a record that is in conflict with a specific
        /// TableOperation. If the operation can start than Locks is updated to reflect the
        /// new TableOperation running, othwerise the function returns false and the manager will
        /// start another Operation making this one wait.
        /// 
        /// </summary>
        /// <param name="o"></param>
        /// <returns></returns>
        private static bool OperationCanStart (DatabaseOperation o) {
            lock (Locks) {
                bool result = true;
                LockRecord opLock = o.GetLockRecord ();

                foreach (LockRecord R in Locks) {
                    if (opLock.DataBaseName == R.DataBaseName) {
                        if (opLock.Schema == R.Schema) {
                            if (opLock.TableName == R.TableName) {
                                if (opLock.PartitionNumber == "") {
                                    result = false;
                                } else {
                                    if (R.PartitionNumber == "") {
                                        result = false;
                                    } else {
                                        if (R.PartitionNumber == opLock.PartitionNumber) {
                                            result = false;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }

                if (result == true) {
                    Locks.Add (opLock);
                }

                return result;
            }
        }

        /// <summary>
        /// 
        /// This static method iterates over all the threads that should run and starts them accordingly to
        /// the MaxNumber passed as a parameter using a semaphore.
        /// 
        /// Please note that we CANNOT use ThreadList.Count instead of RunningThreadCount because we CANNOT
        /// change the ThreadList during thread processing as the foreach iteration runs in paralle with
        /// each single thread and the threads cannot modify the list structure.
        /// 
        /// </summary>
        /// <param name="MaxNumber"></param>
        public static void ProcessAll (int maxNumber) {
            Sem = new Semaphore (maxNumber, maxNumber);                 // Init for the maximum number of parallel threads
            RunningThreadCount = WaitingThreadList.Count;               // Store in RunningThreadCount the number of threads to run
            while (WaitingThreadList.Count > 0) {
                Sem.WaitOne ();                                         // Wait to start
                for (int i = 0; i < WaitingThreadList.Count; i++) {     // Search a Thread to start
                    DatabaseOperation O = WaitingThreadList[i];
                    if (OperationCanStart (O)) {                        // If the thread can safely start
                        WaitingThreadList.Remove (O);                   // Found: remove if from the list of waiting threads
                        Thread runningThread = new Thread (O.Start);    // Initialize a new thread for the operation
                        runningThread.Start ();                         // Starts the thread
                        break;                                          // and finish the loop
                    } else {
                        Logger.LogInfo (String.Format (CultureInfo.InvariantCulture, "{0} : needs to wait.", O.Description));
                    }
                }
            }
            //
            //  We reach this point when all the threads have been detached, some of them will be
            //  still running, so we need to wait for the termination of all the threads
            //
            WorkFinished.WaitOne ();                            
        }

        /// <summary>
        /// 
        /// Adds an operation to the threadlist.
        /// 
        /// </summary>
        /// <param name="T"></param>
        public static void AddOperation (DatabaseOperation databaseOperation) {
            lock (WaitingThreadList) {
                WaitingThreadList.Add (databaseOperation);
            }
        }

        /// <summary>
        /// 
        /// When a thread is finished it has to signal us with this method, so we can decremente the number of
        /// running threads and decide if all the work has been carried on, in this case we Signal on the Event
        /// on which ProcessAll is waiting to make it finish.
        /// 
        /// </summary>
        /// <param name="T"></param>
        public static void ThreadFinished (DatabaseOperation O, bool ExceptionRaised) {
            lock (Locks) {
                foreach (LockRecord R in Locks) {
                    if (R.Operation == O) {
                        Locks.Remove (R);
                        break;
                    }
                }
            }
            Sem.Release ();
            if (ExceptionRaised)
            {
                Interlocked.Increment(ref _exceptionRaisedCount);
            }
            int RunningThread = Interlocked.Decrement (ref RunningThreadCount);
            if (RunningThread == 0) {
                WorkFinished.Set ();
            }
        }
    }

    /// <summary>
    /// 
    /// This class represent an operation that can be carried on by a table or a partition, the real implementation
    /// is in its sub class.
    /// 
    /// </summary>
    public abstract class DatabaseOperation {

        /// <summary>
        /// 
        /// Processes the operation and the signal ThreadManager of its completion
        /// 
        /// </summary>
        public void Start () {
            bool exceptionRaised = true;
            try {
                try {
                    DateTime StartTime = DateTime.Now;
                    Logger.LogInfo (String.Format (CultureInfo.InvariantCulture, "{0} : Started", Description));
                    this.Process ();
                    TimeSpan Elapsed = DateTime.Now - StartTime;
                    Logger.LogInfo (String.Format (CultureInfo.InvariantCulture, "{0} : Completed in {1} seconds", Description, Elapsed.TotalSeconds.ToString ("F", CultureInfo.InvariantCulture)));
                    exceptionRaised = false;
                }
                catch (Exception E)
                {
                    lock (this) {
                        Logger.LogError (String.Format (CultureInfo.InvariantCulture, "{0} thrown an exception ", Description));
                        Logger.LogError (E.Message);
                        Logger.LogError (E.Source);
                        Logger.LogError(E.StackTrace);
                    }
                }
            } finally {
                ThreadManager.ThreadFinished (this, exceptionRaised);
            }
        }

        /// <summary>
        /// 
        /// Returns a LockRecord to manage the handling of locks
        /// 
        /// </summary>
        /// <returns></returns>
        internal abstract ThreadManager.LockRecord GetLockRecord ();

        /// <summary>
        /// 
        /// The real processing is defined by the subclasses, this is only declarative
        /// 
        /// </summary>
        public abstract void Process ();

        /// <summary>
        /// 
        /// Returns a textual description of the operation in progress
        /// 
        /// </summary>
        /// <returns></returns>
        protected abstract string GetDescription ();

        /// <summary>
        /// 
        /// Property that returns the descritption of the database operation
        /// 
        /// </summary>
        public string Description {
            get {
                return GetDescription ();
            }
        }
    }

    /// <summary>
    /// 
    /// Carries on operation on a partition
    /// 
    /// </summary>
    public class PartitionOperation : DatabaseOperation {
        private PartitionCopy P;

        public PartitionOperation (PartitionCopy p) {
            P = p;
        }

        public override void Process () {
            P.Process ();
        }

        protected override string GetDescription () {
            return String.Format (CultureInfo.InvariantCulture, "Partition {0} on {1}.{2} from {3}", P.PartitionId, P.Table.Database.Name, P.Table.DestTableName, P.Table.SourceTableName);
        }

        internal override ThreadManager.LockRecord GetLockRecord () {
            return new ThreadManager.LockRecord (this, P.Table.Database.Name, P.Table.DestTableSchema, P.Table.DestTableName, P.PartitionId);
        }
    }

    /// <summary>
    /// 
    /// Carries on operation on a full table
    /// 
    /// </summary>
    public class TableOperation : DatabaseOperation {
        private TableCopy T;

        public TableOperation (TableCopy pT) {
            T = pT;
        }

        public override void Process () {
            T.ProcessNonPartitionedTable ();
        }

        protected override string GetDescription () {
            return String.Format (CultureInfo.InvariantCulture, "Full table {0}.{1} from {2}", T.Database.Name, T.DestTableName, T.SourceTableName);
        }

        internal override ThreadManager.LockRecord GetLockRecord () {
            return new ThreadManager.LockRecord (this, T.Database.Name, T.DestTableSchema, T.DestTableName);
        }
    }
}
